"""
Catalogs module.
"""

from six import itervalues
from .providers import Provider
from .exceptions import DuplicateKeysError, UndefinedProviderError


GLOBAL_NAMESPACE = '__GLOBAL__'


class Catalog(object):
    """
    Providers catalog.
    """

    namespace = GLOBAL_NAMESPACE
    _providers = None

    def __init__(self, namespace=None):
        """
        Constructor.
        """
        if namespace:
            self.namespace = namespace
        self._providers = dict()
        self._process_attribute_callbacks()

    @property
    def providers(self):
        """
        Providers attribute getter.

        :return:
        """
        for provider in itervalues(self._providers):
            self._update_dependencies_namespace(provider)
        return self._providers

    def bind(self, key, provider):
        """
        Creates key-provider binding.

        :param key:
        :param provider:
        :return:
        """
        if key in self.providers:
            raise DuplicateKeysError('Trying to bind existing key '
                                     '"{0}"'.format(key))
        self._providers[key] = provider
        provider.key = key

    def __setattr__(self, key, value):
        """
        Makes attributes binging.

        :param key:
        :param value:
        :return:
        """
        if isinstance(value, Provider):
            self.bind(key, value)
        else:
            super(Catalog, self).__setattr__(key, value)

    def __getattr__(self, key):
        """
        Returns provider by name.

        :param key:
        :return:
        """
        try:
            return self._providers[key]
        except KeyError:
            raise UndefinedProviderError(
                'Provider "{0}" could not be found'.format(key))

    def _process_attribute_callbacks(self):
        """
        Adds attributes as providers.

        :return:
        """
        for attr_name in set(dir(self.__class__)) - set(dir(Catalog)):
            attr = getattr(self, attr_name)
            if callable(attr) and hasattr(attr, 'provider'):
                key = getattr(attr, 'key', attr_name)

                dependencies = list()
                if hasattr(attr, 'dependencies'):
                    dependencies += attr.dependencies

                self.bind(key, attr.provider(cls=attr,
                                             dependencies=dependencies))
            elif isinstance(attr, Provider):
                self.bind(attr_name, attr)
            else:
                raise UndefinedProviderError('Can not define provider for ' +
                                             repr(attr))

    def _update_dependencies_namespace(self, provider):
        """
        Sets current namespace of all dependencies without namespace.

        :param provider:
        :return:
        """
        if not hasattr(provider, 'dependencies'):
            return
        for require in provider.dependencies:
            if not require.from_namespace:
                require.from_namespace = self.namespace
